/*
 *  grppasswd utils
 * 
 *  Copyright (C) 1997  Ragnar Hojland Espinosa
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *  You may contact the author for ideas, comments or bug reports at
 * 
 *                                    tech.support@redestb.es
 *                                    ragnar@dfyre.ccon.net
 */


#include <sys/file.h>
#include <stdio.h>
#include <errno.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/signal.h>

#include "common.h"

bool lock_file (FILE *fp)
  {
    if (!fp || flock (fileno(fp), LOCK_EX) == -1)
      {
	perror ("Error");
	return false;
      }
    
    return true;
  }

bool unlock_file (FILE *fp)
  {
    if (!fp || flock (fileno(fp), LOCK_UN) == -1)
      {
	perror ("Error");
	return false;
      }
    
    return true;
  }

void prompt_text (const char *prompt, char *dest, size_t length)
  {
    printf (prompt);
    fgets (dest, length, stdin);
    if (dest[0] != '\0')
      {
	dest[strlen(dest)-1] = '\0';
      }
  }



/* this comes from passwd.c, if I remember right */
void set_max_resources (void)
  {
    struct rlimit rlim;
    
    rlim.rlim_cur = rlim.rlim_max = RLIM_INFINITY;
    setrlimit (RLIMIT_CPU,   &rlim);
    setrlimit (RLIMIT_FSIZE, &rlim);
    setrlimit (RLIMIT_STACK, &rlim);
    setrlimit (RLIMIT_DATA,  &rlim);
    setrlimit (RLIMIT_RSS,   &rlim);
    rlim.rlim_cur = rlim.rlim_max = 0;
    setrlimit (RLIMIT_CORE,  &rlim);
    
    signal (SIGALRM, SIG_IGN);
    signal (SIGHUP,  SIG_IGN);
    signal (SIGINT,  SIG_IGN);
    signal (SIGPIPE, SIG_IGN);
    signal (SIGQUIT, SIG_IGN);
    signal (SIGTERM, SIG_IGN);
    signal (SIGTSTP, SIG_IGN);
    signal (SIGTTOU, SIG_IGN);

    /* this doesnt */
    umask (077);
  }

