/*
 *  grppasswd utils
 * 
 *  Copyright (C) 1997  Ragnar Hojland Espinosa
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *  You may contact the author for ideas, comments or bug reports at
 * 
 *                                    tech.support@redestb.es
 *                                    ragnar@dfyre.ccon.net
 */


#include <unistd.h>
#include <stdio.h>
#include <grp.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <getopt.h>
#include <stdlib.h>
#include <sys/syslog.h>
#include "common.h"
#include "grppasswd.h"

/* 
 * A little effort to write it, a great panic after overwriting it when it was
 * done, and the idea of making a curses based fs editor after recovering it :)
 *
 * Added addgroup later in the same program, which explains the not so optimal
 * structure and symmetry. I'll clean it soon'
*/

int main (int argc, char *argv[])
  {   
    openlog ("grppasswd", 0, LOG_AUTHPRIV);
    
    if (strstr (argv[0], "grppasswd"))
      {    
	struct group *g;
	char password[0xff];

	if (argc > 3)
	  {
	    fputs ("Too many arguments\n", stderr);
	    return 1;
	  }
	
	if (argc == 1)
	  {
	    g = getgroup ("Group: ", 0);
	  }
	else
	  {
	    g = getgroup (0, argv[1]);
	  }
	
	if (!g)
	  {
	    fputs ("Unknown group\n", stderr);
	    return 1;
	  }
	
#ifdef GROUP_LEADER_SUPPORT
	if (getuid() != 0)
	  {	    
	    if (!g->gr_mem || !g->gr_mem[0] || g->gr_mem[0][0] == '\0' ||
		strcmp (g->gr_mem[0], getlogin()) )
	      {
		fputs ("Only superuser or group leader may change group password\n", stderr);
		return 1;
	      }
	  }
#endif	
	
	if (argc != 1)
	  {
	    printf ("Changing password for group %s\n", g->gr_name);
	  }


#ifdef GROUP_LEADER_SUPPORT
	if (getuid() != 0 && g->gr_passwd)
	  {
	    char buffer[0xff], salt[2];
	    strcpy (buffer, getpass ("Enter old password: "));
	    salt[0] = g->gr_passwd[0];
	    salt[1] = g->gr_passwd[1];
	    if (strcmp (crypt (buffer, salt), g->gr_passwd))
	      {
		syslog (LOG_NOTICE, 
			"%s failed to enter password for group %s.",
			getlogin()?getlogin():"UNKNOWN", g->gr_name);
		fputs ("Invalid password.\n", stderr);
		return 1;
	      }
	  }
#endif	
	
	if (argc == 1)
	  {
	    prompt_password (0, password, sizeof(password));
	  }
	else
	  {
	    prompt_password (argv[2], password, sizeof(password));
	  }
	    
	if (password[0] == '\0' && argc < 3)
	  {
	    puts ("Password not changed.");
	    return 0;
	  }
	
	if (!valid_password (password, g))
	  {
	    fputs ("Invalid password.\n", stderr);
	    return 1;
	  }
	
	if (!password_update (password, g->gr_gid))
	  {
	    fprintf (stderr, "Error updating %s\n", GROUP_FILE);
	    return 1;
	  }

	syslog (LOG_NOTICE, "%s changed password for group %s.",
		getlogin()?getlogin():"UNKNOWN", g->gr_name);
	puts ("Group password changed.");    
	return 0;
      }
    
    else if (strstr (argv[0], "addgroup"))
      {		
	struct group *g, newgrp;
	int o, longindex;
	char  groupname[0xff];
	char  grouppass[0xff];
	char  groupuser[0xfff];
	char  *dummy[2];
	gid_t groupgid;
	bool  groupgid_set;
	bool  grouppass_set;
	bool  groupuser_set;
	
	const struct option longopts[] = 
	  {
	    { "name",     required_argument, 0, 'n' },
	    { "gid",      required_argument, 0, 'g' },
	    { "password", required_argument, 0, 'p' },
	    { "users",    required_argument, 0, 'u' }, 
	    { "help",     no_argument,       0, 'h' },
	    { "version",  no_argument,       0, 'V' },
	    { 0, 0, 0, 0 }
	  };

	if (getuid() != 0)
	  {
	    fputs ("Only superuser may run addgroup, sorry\n", stderr);
	    return 1;
	  }
	
	groupname[0]  = '\0';
	grouppass[0]  = '\0';
	groupuser[0]  = '\0';
	groupgid_set  = false;
	grouppass_set = false;
	groupuser_set = false;
	
	while ( (o=getopt_long (argc, argv, "hVn:g:p:", longopts, &longindex)) != -1)
        switch (o)
	  {
	    case 'h':
	      addgroup_showhelp(argv[0]);
	      return 0;
	    
	    case 'V':
	      addgroup_showversion();
	      return 0;
	    
	    case 'n':
	      strncpy (groupname, optarg, 0xff);
	      break;
	    
	    case 'g':
	      groupgid_set = true;
	      groupgid = atoi (optarg);
	      break;
	    
	    case 'p':
	      grouppass_set = true;
	      strncpy (grouppass, optarg, 0xff);
	      break;

	    case 'u':
	      strncpy (groupuser, optarg, 0xfff);
	      break;
	    
	    default:
	      fprintf (stderr, "Try `%s --help' for more information.\n",
		       argv[0]);
	      return 1;
	    
	  }

	
	if (groupname[0] == '\0')
	  {
	    puts ("\n"
		  "Adding a new group. The group name should not exceed 8 characters\n"
		  "in length, or you may run into problems later.\n");

	    prompt_text ("New group name: ", groupname, 0xff);
	  }

	if (groupname[0] == '\0')
	  {
	    puts ("Cancelled.");
	    return 0;
	  }
	
	g = getgroup (0, groupname);

	if (g)
	  {
	    fputs ("That group already exists.\n", stderr);
	    return 1;
	  }

	if (!groupgid_set)
	  {
	    groupgid = find_freegid();
	    prompt_gid (&groupgid);
	  }
	
	if (getgrgid (groupgid))
	  {
	    fputs ("That gid already exists.\n", stderr);
	    return 1;
	  }

	
	if (!groupuser_set)
	  {
	    prompt_text ("Enter user list: ", &groupuser[0], 0xfff);
	  }
	
	if (!grouppass_set)
	  {
	    if (!prompt_password (0, grouppass, 0xff))
	      {
		puts ("Group not added.");
		return 1;
	      }
	  }
	       	
	dummy[0] = &groupuser[0];
	dummy[1] = 0;
	newgrp.gr_name   = &groupname[0];
	newgrp.gr_passwd = &grouppass[0];
	newgrp.gr_gid    = groupgid;
	newgrp.gr_mem    = &dummy[0];
	
	
	if (!valid_password (grouppass, &newgrp))
	  {
	    fputs ("Invalid password.\n", stderr);
	    return 1;
	  }
	
	
	if (!addgroup_update (&newgrp))
	  {
	    fprintf (stderr, "Error updating %s\n", GROUP_FILE);
	    return 1;
	  }
	
	syslog (LOG_USER, "New group %s added by %s.", 
		newgrp.gr_name, getlogin()?getlogin():"UNKNOWN");
	puts ("New group added.");
	return 0;
      }
    
    else
      {
	fputs ("Program should be called addgroup or grppasswd.", stderr);
	return 1;
      }

    
    return 0;
  }



struct group* getgroup (const char *prompt, const char *name)
  {
    char buffer[0xff];
    
    if (!name)
      {
	prompt_text (prompt, buffer, 0xff);
	if (buffer[0] == '\0')
	  {
	    return 0;
	  }
	
	return getgrnam (buffer);
      }
    
    return getgrnam (name);
  }


bool prompt_password (const char *password, char *new_password, const size_t length)
  {
    char buffer[0xff];

    if (!password)
      {
	strncpy (new_password, getpass("Enter new password: "), 0xff);
	new_password[MAX_PASSWORD_LENGTH] = '\0';
	
	if (new_password[0] == '\0')
	  {
	    return true;
	  }
	
    	strncpy (buffer, getpass("Re-type new password: "), 0xff);
	buffer[MAX_PASSWORD_LENGTH] = '\0';
	
	if (strcmp (new_password, buffer))
	  {
	    printf ("You misspelled it. ");
	    new_password[0] = '\0';
	    return false;
	  }
	
	return true;
      }
    
    buffer[MAX_PASSWORD_LENGTH] = 0;
    strncpy (new_password, password, length);    
    return true;
  }


/* Since the only one who is going to be able to change the group password
 * is root, validity is extremely relaxed */
bool valid_password (const char *password, const struct group *grp)
  {
    if (!strcmp (password, grp->gr_name))
      {
	return false;
      }
    
    return true;
  }


bool password_update (const char *password, const gid_t gid)
  {
    FILE *out;
    struct group *g;
    char salt[2];
    time_t tm;
    
    set_max_resources();
    
    out = fopen (TMP_FILE, "w");
    if (!out)
      {
	int e = errno;
	fprintf (stderr, "Error while opening %s: %s\n",
		 TMP_FILE, strerror(e));
	return false;
      }    

    if (!lock_file(out))
      {
	fprintf (stderr, "Couldn't lock %s\n", TMP_FILE);
	return false;
      }
    
    if (chmod (TMP_FILE, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) == -1)
      {
	perror ("chmod: ");
	return false;
      }
    
    time (&tm);
    salt[0] = btoa (tm & 0x3f);
    salt[1] = btoa ((tm >> 6) & 0x3f);
    
    setgrent();
    
    while ( (g = getgrent()))
      {
	int i;
	fprintf (out, "%s:%s:%u:", 
		 g->gr_name, 
		 g->gr_gid == gid ? 
		     password[0] == '\0' ? "" 
		                           : crypt (password, salt) 
		     : g->gr_passwd, 
		 g->gr_gid);
	for (i = 0; g->gr_mem[i]; i++)
	  {
	    if (i)
	      {
		fputc (',', out);
	      }
	    fprintf (out, "%s", g->gr_mem[i]);
	  }
	fputc ('\n', out);
      }
    
    if (errno == ENOMEM)
      {
	perror ("Error");
	endgrent();
	fclose (out);
	remove (TMP_FILE);
	return false;
      }
    
    endgrent();    
    rename (TMP_FILE, GROUP_FILE);
    
    if (!unlock_file(out))
      {
	fprintf (stderr, "Couldn't unlock %s\n", TMP_FILE);
	fclose (out);
	return false;
      }

    fclose (out);    
    return true;
  }

void addgroup_showhelp (const char *progname)
  {
    printf ("Usage: %s [OPTION]\n\n"
	    "  -n, --name=WORD\n"
	    "  -g, --gid=GID\n"
	    "  -p, --password=WORD\n"
	    "  -u, --users=WORD\n"
	    "  -h, --help\n"
	    "  -V, --version\n",
	    progname);
  }

void addgroup_showversion (void)
  {
    printf (GRPPASSWD_VERSION_INFO
	    "Compiled: %s - Group leader support "
#ifdef GROUP_LEADER_SUPPORT
	    "enabled"
#else
	    "disabled"
#endif
	    "\n", 
	    __DATE__
 );
  }

gid_t find_freegid (void)
  {
    gid_t gid = 0;

    while (getgrgid (++gid))
      { }
    
    return gid;
  }

void prompt_gid (gid_t *gid)
  {
    char buffer[0xff];
    
    printf ("Enter new gid [%u]: ", *gid);
    fgets (buffer, 0xff, stdin);
    if (buffer[0] == '\n')
      {
	return;
      }

    *gid = atoi (buffer);
  }


bool addgroup_update (const struct group *grp)
  {
    FILE *out;
    char salt[2];
    time_t tm;
    
    set_max_resources();

    out = fopen (GROUP_FILE, "a");
    if (!out)
      {
	int e = errno;
	fprintf (stderr, "Error while opening %s: %s\n",
		 GROUP_FILE, strerror(e));
	return false;
      }    
    
    if (!lock_file(out))
      {
	fprintf (stderr, "Couldn't lock %s\n", GROUP_FILE);
	return false;
      }
    
    time (&tm);
    salt[0] = btoa (tm & 0x3f);
    salt[1] = btoa ((tm >> 6) & 0x3f);

    fprintf (out, "%s:%s:%u:%s\n",
		 grp->gr_name, 	       
		 (grp->gr_passwd[0] == '\0' ?
		    "" : crypt (grp->gr_passwd, salt) ),
		 grp->gr_gid, grp->gr_mem[0]);
    
    if (!unlock_file(out))
      {
	fprintf (stderr, "Couldn't unlock %s\n", GROUP_FILE);
	fclose (out);
	return false;
      }

    fclose (out);        
    return true;    
  }


